
/*
 *  arch/i386/kernel/rt_time.c
 *
 *  Copyright (C) 1996 by Michael Barabanov
 *
 */


#include <asm/io.h>
#include <linux/errno.h>
#include <asm/system.h>
#include <asm/rt_time.h>
#include <asm/rt_irq.h>
#include <linux/sched.h>
#include <linux/timex.h>

#ifdef __RT_DEBUG__
#include <linux/cons.h>
#endif

/* this is a fake version
   1. steal the interrupt
   2. reset the clock to make timer 0 periodic. 
   3. 

   

extern unsigned long SFREQ;

static RTIME rt_linux_clock;    /* time to generate the next linux interrupt  */
static RTIME rt_callback_time;  /* time to invoke the supplied procedure      */

static RTIME rt_intr_time;

static void (*rt_timer_handler)(void);


#define READ_CNT0(var) \
{ var = inb(0x40); var |= (inb(0x40) << 8); }

#define READ_CNT2(var) \
{ var = inb(0x42); var |= (inb(0x42) << 8); }


#define LATCH_CNT0() \
outb(0xd2, 0x43);

#define LATCH_CNT0_AND_2() \
outb(0xda, 0x43);

#define LATCH_CNT2() \
outb(0xd8, 0x43);


static RTIME rt_base_time;      /* the time when Counter 0 was == base_c0 */
static int base_c0;

int rt_latch2;			/* the current latch value for counter 2 */
                                

RTIME rt_get_time(void)
{
	register int c0;
	int flags;
	RTIME result;

	r_save_flags(flags);
	r_cli();
	LATCH_CNT0();
	READ_CNT0(c0);
	result = rt_base_time + ((base_c0 - c0) & 0xffff);
	r_restore_flags(flags);
	return result;
}


#ifdef __RT_DEBUG__
int maxdiff, mindiff;
#endif

/* The main procedure; resets the 8254 timer to generate an interrupt.  The
 * tricky part is to keep the global time while reprogramming it.  We latch
 * counters 0 and 2 atomically before and after reprogramming to figure it out.
 */

static void rt_reset_timer(void)
{
	register int c2;
	register int c2new;
	register int temp;
	int incr;
	int flags;
	

#ifdef __RT_DEBUG__
	int diff;
#endif

	r_save_flags(flags);
	r_cli();

	if (rt_linux_clock < rt_callback_time) {
		temp = rt_linux_clock - rt_base_time;
	}
	else {
		temp = rt_callback_time - rt_base_time;
	}

	/* between the two latches we want to have as less stuff as possible */
	LATCH_CNT0_AND_2();
	READ_CNT0(c2new);
	READ_CNT2(c2);

	incr = ((base_c0 - c2new) & 0xffff);  /* current time = rt_base_time + incr */

	temp -= incr;
	if (temp < 10) {
		temp = 10; 
	}

	outb(temp & 0xff, 0x40);
	outb_p(temp >> 8, 0x40);
    
	LATCH_CNT0_AND_2();
	READ_CNT0(base_c0);
	READ_CNT2(c2new);

	rt_base_time += (incr + ((c2new < c2) ? (c2 - c2new) / 2 : (c2 - c2new + rt_latch2) / 2));

	rt_intr_time = rt_base_time + temp;
	

#ifdef __RT_DEBUG__
	diff = (c2new < c2) ? (c2 - c2new) / 2 : (c2 - c2new + rt_latch2) / 2;
	if (diff > maxdiff) {
		maxdiff = diff;
		conpr("new record: "); conprn(mindiff); conprn(maxdiff); conpr("\n");
	}
	if (diff < mindiff) {
		mindiff = diff;
		conpr("new record: "); conprn(mindiff); conprn(maxdiff); conpr("\n");
	}
	r_restore_flags(flags);
#endif
}


void rt_timer_irq(void)
{
	if (rt_intr_time + 10 >= rt_callback_time) {
		(*rt_timer_handler)();

	} else {
		SFREQ |= 1;
		rt_linux_clock += LATCH;
		rt_reset_timer();
	}

}


void rt_set_timer(RTIME when) 
{
	rt_callback_time = when;
	rt_reset_timer();
}


void rt_no_timer(void) 
{
	rt_callback_time = RT_TIME_END;
	rt_reset_timer();
}


int rt_request_timer(void (*fn)(void)) 
{
	if (!fn) {
		return -EINVAL;
	}
	if (rt_timer_handler) {
		return -EBUSY;
	}

	rt_timer_handler = fn;
	return 0;
}


void rt_free_timer(void)
{
	rt_timer_handler = 0;
	rt_no_timer();
}



void rt_time_init(void)
{
	int flags;

	r_save_flags(flags);
	r_cli();

	rt_timer_handler = 0;

	rt_linux_clock = LATCH;
	rt_callback_time = RT_TIME_END;

	outb_p(0x30, 0x43);       /* 8254, channel 0, mode 0, lsb+msb */

	request_RTirq(0, rt_timer_irq);

	outb_p(0xff, 0x40);       /* start counting */
	outb_p(0xff, 0x40);

	rt_base_time = 0;
	LATCH_CNT0();
	READ_CNT0(base_c0);

	rt_reset_timer();

#ifdef __RT_DEBUG__
	mindiff = 10000;
	maxdiff = -10000;
#endif /* __RT_DEBUG__ */

	/* program channel 2 of the 8253 chip for periodic interrrupt */
	rt_latch2 = 50000;
	outb_p(0xb6, 0x43);	/* binary, mode 3, LSB/MSB, ch 2 */
	outb_p(rt_latch2 & 0xff, 0x42);
	outb((rt_latch2 >> 8) & 0xff, 0x42);

	outb((inb_p(0x61) & 0xfd) | 1, 0x61); /* shut up the speaker and enable counting */

	r_restore_flags(flags);
}

